@@ -1,6 +1,6 @@
 /* dvipsencoding.{cc,hh} -- store a DVIPS encoding
  *
- * Copyright (c) 2003-2005 Eddie Kohler
+ * Copyright (c) 2003-2007 Eddie Kohler
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -27,7 +27,8 @@
 #include "util.hh"
 
 static String::Initializer initializer;
-enum { GLYPHLIST_MORE = 0x40000000, U_EMPTYSLOT = 0xD801 };
+enum { GLYPHLIST_MORE = 0x40000000,
+       U_EMPTYSLOT = 0xD801, U_ALTSELECTOR = 0xD802 };
 static HashMap<String, int> glyphlist(-1);
 static PermString::Initializer perm_initializer;
 PermString DvipsEncoding::dot_notdef(".notdef");
@@ -423,7 +424,7 @@ DvipsEncoding::parse_position_words(Vector<String> &v, int override, ErrorHandle
     if (v.size() != 4)
 	return -EPARSE;
 
-    int c = encoding_of(v[0], override > 0);
+    int c = encoding_of(v[0]);
     if (c < 0)
 	return (override > 0 ? errh->warning("'%s' has no encoding, ignoring positioning", v[0].c_str()) : -1);
     
@@ -451,7 +452,7 @@ DvipsEncoding::parse_unicoding_words(Vector<String> &v, int override, ErrorHandl
     if (v.size() < 2 || (v[1] != "=" && v[1] != "=:" && v[1] != ":="))
 	return -EPARSE;
     else if (v[0] == "||" || (av = encoding_of(v[0])) < 0)
-	return errh->error("target '%s' has no encoding, ignoring UNICODING", v[0].c_str());
+	return errh->warning("'%s' has no encoding, ignoring UNICODING", v[0].c_str());
 
     int original_size = _unicoding.size();
     
@@ -689,7 +690,7 @@ DvipsEncoding::bad_codepoint(int code, Metrics &metrics, Vector<String> &unencod
 	    Vector<Setting> v;
 	    v.push_back(Setting(Setting::RULE, 500, 500));
 	    v.push_back(Setting(Setting::SPECIAL, String("Warning: missing glyph '") + _e[code] + "'"));
-	    metrics.encode_virtual(code, _e[code], 0, v);
+	    metrics.encode_virtual(code, _e[code], 0, v, true);
 	    unencoded.push_back(_e[code]);
 	}
     }
@@ -730,16 +731,12 @@ DvipsEncoding::x_unicodes(PermString chname, Vector<uint32_t> &unicodes) const
 
 
 void
-DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap, Efont::Cff::Font *font, Secondary *secondary, bool literal, ErrorHandler *errh)
+DvipsEncoding::make_metrics(Metrics &metrics, const FontInfo &finfo, Secondary *secondary, bool literal, ErrorHandler *errh)
 {
     // first pass: without secondaries
     for (int code = 0; code < _e.size(); code++) {
 	PermString chname = _e[code];
 
-	// the altselector character has its own glyph name
-	if (code == _altselector_char && !literal)
-	    chname = "altselector";
-
 	// common case: skip .notdef
 	if (chname == dot_notdef)
 	    continue;
@@ -752,13 +749,11 @@ DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap,
 	Efont::OpenType::Glyph glyph = 0;
 	uint32_t glyph_uni = (unicodes.size() ? unicodes[0] : 0);
 	for (uint32_t *u = unicodes.begin(); u < unicodes.end() && !glyph; u++)
-	    if ((glyph = map_uni(*u, cmap, metrics)) > 0)
+	    if ((glyph = map_uni(*u, *finfo.cmap, metrics)) > 0)
 		glyph_uni = *u;
 
 	// find named glyph, if any
-	Efont::OpenType::Glyph named_glyph = 0;
-	if (font)
-	    named_glyph = font->glyphid(chname);
+	Efont::OpenType::Glyph named_glyph = finfo.glyphid(chname);
 
 	// do not use a Unicode-mapped glyph if literal
 	if (literal)
@@ -782,18 +777,12 @@ DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap,
 	
 	PermString chname = _e[code];
 
-	// the altselector character has its own glyph name
-	if (code == _altselector_char && !literal)
-	    chname = "altselector";
-
 	// find all Unicodes
 	Vector<uint32_t> unicodes;
 	bool unicodes_explicit = x_unicodes(chname, unicodes);
 
 	// find named glyph, if any
-	Efont::OpenType::Glyph named_glyph = 0;
-	if (font)
-	    named_glyph = font->glyphid(chname);
+	Efont::OpenType::Glyph named_glyph = finfo.glyphid(chname);
 
 	// 1. We were not able to find the glyph using Unicode.
 	// 2. There might be a named_glyph.
@@ -818,6 +807,13 @@ DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap,
 	/* all set */;
     }
 
+    // add altselector
+    if (_altselector_char >= 0 && _altselector_char < _e.size()) {
+	metrics.add_altselector_code(_altselector_char, 0);
+	if (metrics.glyph(_altselector_char) <= 0 && !literal)
+	    (void) secondary->encode_uni(_altselector_char, "altselector", U_ALTSELECTOR, metrics, errh);
+    }
+
     // final pass: complain
     Vector<String> unencoded;
     for (int code = 0; code < _e.size(); code++)
@@ -842,6 +838,27 @@ DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap,
     metrics.set_coding_scheme(_coding_scheme);
 }
 
+void
+DvipsEncoding::make_base_mappings(Vector<int> &mappings, const FontInfo &finfo)
+{
+    mappings.clear();
+    for (int code = 0; code < _e.size(); code++) {
+	PermString chname = _e[code];
+
+	// common case: skip .notdef
+	if (chname == dot_notdef)
+	    continue;
+
+	// find named glyph
+	Efont::OpenType::Glyph named_glyph = finfo.glyphid(chname);
+	if (named_glyph > 0) {
+	    if (mappings.size() <= named_glyph)
+		mappings.resize(named_glyph + 1, -1);
+	    mappings[named_glyph] = code;
+	}
+    }
+}
+
 
 void
 DvipsEncoding::apply_ligkern_lig(Metrics &metrics, ErrorHandler *errh) const
