@@ -1,6 +1,6 @@
 /* metrics.{cc,hh} -- an encoding during and after OpenType features
  *
- * Copyright (c) 2003-2006 Eddie Kohler
+ * Copyright (c) 2003-2007 Eddie Kohler
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -169,13 +169,15 @@ Metrics::encode(Code code, uint32_t uni, Glyph g)
 }
 
 void
-Metrics::encode_virtual(Code code, PermString name, uint32_t uni, const Vector<Setting> &v)
+Metrics::encode_virtual(Code code, PermString name, uint32_t uni, const Vector<Setting> &v, bool base_char)
 {
     assert(code >= 0 && v.size() > 0);
     if (code >= _encoding.size())
 	_encoding.resize(code + 1, Char());
     _encoding[code].unicode = uni;
     _encoding[code].glyph = VIRTUAL_GLYPH;
+    if (base_char)
+	_encoding[code].flags |= Char::BASE_REP;
     assert(!_encoding[code].virtual_char);
     VirtualChar *vc = _encoding[code].virtual_char = new VirtualChar;
     vc->name = name;
@@ -188,6 +190,25 @@ Metrics::encode_virtual(Code code, PermString name, uint32_t uni, const Vector<S
     }
 }
 
+void
+Metrics::apply_base_encoding(const String &font_name, const DvipsEncoding &dvipsenc, const Vector<int> &mapping)
+{
+    int font_number = -1;
+    for (Char *c = _encoding.begin(); c != _encoding.end(); c++)
+	if (c->glyph > 0 && !c->virtual_char && c->glyph < mapping.size()
+	    && mapping[c->glyph] >= 0) {
+	    if (font_number < 0)
+		font_number = add_mapped_font(mapped_font(0), font_name);
+	    VirtualChar *vc = c->virtual_char = new VirtualChar;
+	    vc->name = dvipsenc.encoding(mapping[c->glyph]);
+	    vc->setting.push_back(Setting(Setting::FONT, font_number));
+	    vc->setting.push_back(Setting(Setting::SHOW, mapping[c->glyph], c->glyph));
+	    c->glyph = VIRTUAL_GLYPH;
+	    c->base_code = -1;
+	    c->flags = (c->flags & ~Char::BASE_LIVE) | Char::BASE_REP;
+	}
+}
+
 void
 Metrics::add_altselector_code(Code code, int altselector_type)
 {
@@ -199,13 +220,17 @@ Metrics::add_altselector_code(Code code, int altselector_type)
     _altselectors.push_back(Kern(code, altselector_type));
 }
 
-void
-Metrics::base_glyphs(Vector<Glyph> &v) const
+bool
+Metrics::base_glyphs(Vector<Glyph> &v, int size) const
 {
+    bool any = false;
     v.assign(_encoding.size(), 0);
-    for (Code c = 0; c < _encoding.size(); c++)
-	if (_encoding[c].base_code >= 0)
-	    v[_encoding[c].base_code] = _encoding[c].glyph;
+    for (const Char *ch = _encoding.begin(); ch != _encoding.end(); ch++)
+	if (ch->base_code >= 0 && ch->base_code < size) {
+	    v[ch->base_code] = ch->glyph;
+	    any = true;
+	}
+    return any;
 }
 
 
@@ -817,10 +842,17 @@ Metrics::all_ligatures(Vector<Ligature3> &all_ligs) const
 }
 
 void
-Metrics::mark_liveness(int size, const Vector<Ligature3> &all_ligs)
+Metrics::mark_liveness(int size, const Vector<Ligature3> *all_ligs)
 {
     _liveness_marked = true;
     bool changed;
+
+    // make sure we have ligatures
+    Vector<Ligature3> my_ligs;
+    if (!all_ligs) {
+	all_ligatures(my_ligs);
+	all_ligs = &my_ligs;
+    }
     
     /* Characters below 'size' are in both virtual and base encodings. */
     for (Char *ch = _encoding.begin(); ch < _encoding.begin() + size; ch++)
@@ -829,7 +861,7 @@ Metrics::mark_liveness(int size, const Vector<Ligature3> &all_ligs)
 
     /* Characters reachable from live chars by live ligatures are live. */
   redo_live_reachable:
-    for (const Ligature3 *l = all_ligs.begin(); l != all_ligs.end(); l++)
+    for (const Ligature3 *l = all_ligs->begin(); l != all_ligs->end(); l++)
 	if (_encoding[l->in1].flag(Char::LIVE) && _encoding[l->in2].flag(Char::LIVE)) {
 	    Char &ch = _encoding[l->out];
 	    if (!ch.flag(Char::LIVE))
@@ -859,7 +891,8 @@ Metrics::mark_liveness(int size, const Vector<Ligature3> &all_ligs)
 	    if (VirtualChar *vc = ch->virtual_char) {
 		int font_number = 0;
 		for (Setting *s = vc->setting.begin(); s != vc->setting.end(); s++)
-		    if (s->op == Setting::SHOW && font_number == 0)
+		    if (s->op == Setting::SHOW && font_number == 0
+			&& _encoding[s->x].base_code >= 0)
 			_encoding[s->x].flags |= Char::BASE_LIVE;
 		    else if (s->op == Setting::FONT)
 			font_number = s->x;
@@ -917,8 +950,12 @@ Metrics::cut_encoding(int size)
 
     /* Change "emptyslot"s to ".notdef"s. */
     for (Char *ch = _encoding.begin(); ch != _encoding.end(); ch++)
-	if (ch->glyph == emptyslot_glyph())
-	    ch->glyph = 0, ch->base_code = -1;
+	if (ch->glyph == emptyslot_glyph()) {
+	    ch->glyph = 0;
+	    ch->base_code = -1;
+	    // 21.Feb.2007: Character isn't live any more.
+	    ch->flags &= ~(Char::BASE_LIVE | Char::LIVE);
+	}
     
     /* Maybe we don't need to do anything else. */
     if (_encoding.size() <= size) {
@@ -927,11 +964,8 @@ Metrics::cut_encoding(int size)
     }
 
     /* Need liveness markings. */
-    if (!_liveness_marked) {
-	Vector<Ligature3> all_ligs;
-	all_ligatures(all_ligs);
-	mark_liveness(size, all_ligs);
-    }
+    if (!_liveness_marked)
+	mark_liveness(size);
     
     /* Characters below 'size' are 'good'.
        Characters above 'size' are not 'good'. */
@@ -1051,7 +1085,7 @@ Metrics::shrink_encoding(int size, const DvipsEncoding &dvipsenc, ErrorHandler *
 
     /* Need liveness markings. */
     if (!_liveness_marked)
-	mark_liveness(size, all_ligs);
+	mark_liveness(size, &all_ligs);
     
     /* Score characters by importance. Importance relates first to Unicode
        values, and then recursively to the importances of characters that form
@@ -1202,28 +1236,21 @@ font, so I've ignored these:\n%s.)", sa.c_str());
 void
 Metrics::make_base(int size)
 {
-    if (_encoding.size() <= size)
-	/* If the encoding has not grown, there cannot be any base characters
-	   left to swap in. */
-	return;
-    
-    assert(_encoding.size() > size);
-    bool reencoded = false;
     Vector<Code> reencoding;
-    for (Code c = 0; c < _encoding.size(); c++)
-	reencoding.push_back(c);
-    for (Code c = 0; c < size; c++) {
+    for (Code c = 0; c < size && c < _encoding.size(); c++) {
 	Char &ch = _encoding[c];
 	if (ch.base_code >= 0 && ch.base_code != c) {
+	    if (!reencoding.size())
+		for (Code cc = 0; cc < _encoding.size(); cc++)
+		    reencoding.push_back(cc);
 	    reencoding[ch.base_code] = c;
 	    reencoding[c] = ch.base_code;
 	    _encoding[c].swap(_encoding[ch.base_code]);
-	    reencoded = true;
 	}
 	if (ch.virtual_char)	// remove it
 	    ch.clear();
     }
-    if (reencoded) {
+    if (reencoding.size()) {
 	reencode(reencoding);
 	cut_encoding(size);
     }
@@ -1240,7 +1267,19 @@ Metrics::need_virtual(int size) const
     if (size > _encoding.size())
 	size = _encoding.size();
     for (const Char *ch = _encoding.begin(); ch < _encoding.begin() + size; ch++)
-	if (ch->pdx || ch->pdy || ch->adx || ch->virtual_char)
+	if (ch->glyph /* actually encoded */
+	    && (ch->pdx || ch->pdy || ch->adx || ch->virtual_char))
+	    return true;
+    return false;
+}
+
+bool
+Metrics::need_base()
+{
+    if (!_liveness_marked)
+	mark_liveness(_encoding.size());
+    for (const Char *ch = _encoding.begin(); ch < _encoding.end(); ch++)
+	if ((ch->flags & Char::BASE_LIVE) && ch->glyph != _boundary_glyph)
 	    return true;
     return false;
 }
@@ -1248,6 +1287,8 @@ Metrics::need_virtual(int size) const
 bool
 Metrics::setting(Code code, Vector<Setting> &v, SettingMode sm) const
 {
+    extern int letterspace;
+    
     if (!(sm & SET_KEEP))
 	v.clear();
     
@@ -1259,6 +1300,10 @@ Metrics::setting(Code code, Vector<Setting> &v, SettingMode sm) const
     if (const VirtualChar *vc = ch.virtual_char) {
 	bool good = true;
 	int font_number = 0;
+	
+	if (ch.pdx != 0 || ch.pdy != 0)
+	    v.push_back(Setting(Setting::MOVE, ch.pdx, ch.pdy));
+	
 	for (const Setting *s = vc->setting.begin(); s != vc->setting.end(); s++)
 	    switch (s->op) {
 	      case Setting::MOVE:
@@ -1279,16 +1324,24 @@ Metrics::setting(Code code, Vector<Setting> &v, SettingMode sm) const
 		    v.push_back(*s);
 		break;
 	      case Setting::KERN:
+	      case Setting::KERNX:
 		if (sm & SET_INTERMEDIATE)
 		    v.push_back(*s);
-		else if (s > vc->setting.begin() && s + 1 < vc->setting.end()
+		else if (font_number == 0 && s > vc->setting.begin()
+			 && s + 1 < vc->setting.end()
 			 && s[-1].op == Setting::SHOW
-			 && s[1].op == Setting::SHOW
-			 && font_number == 0)
-		    if (int k = kern(s[-1].x, s[1].x))
+			 && s[1].op == Setting::SHOW) {
+		    int k = kern(s[-1].x, s[1].x);
+		    if (s->op == Setting::KERNX)
+			k -= letterspace;
+		    if (k)
 			v.push_back(Setting(Setting::MOVE, k, 0));
+		}
 		break;
 	    }
+
+	if (ch.pdy != 0 || ch.adx - ch.pdx != 0)
+	    v.push_back(Setting(Setting::MOVE, ch.adx - ch.pdx, -ch.pdy));
 	return good;
 	
     } else if (ch.base_code >= 0) {
@@ -1355,56 +1408,69 @@ Metrics::kerns(Code in1, Vector<Code> &in2, Vector<int> &kern) const
 /*****************************************************************************/
 /* debugging								     */
 
+void
+Metrics::unparse(const Char *ch) const
+{
+    Code c;
+    if (ch >= _encoding.begin() && ch < _encoding.end())
+	c = ch - _encoding.begin();
+    else
+	c = -1;
+    fprintf(stderr, "%4d/%s%s%s%s%s%s\n", c, code_str(c),
+	    (ch->flag(Char::LIVE) ? " [L]" : ""),
+	    (ch->flag(Char::BASE_LIVE) ? " [B]" : ""),
+	    (ch->flag(Char::CONTEXT_ONLY) ? " [C]" : ""),
+	    (ch->flag(Char::BUILT) ? " [!]" : ""),
+	    (ch->base_code >= 0 ? " <BC>" : ""));
+    if (ch->base_code >= 0 && ch->base_code != c)
+	fprintf(stderr, "\tBASE %d/%s\n", ch->base_code, code_str(ch->base_code));
+    if (const VirtualChar *vc = ch->virtual_char) {
+	fprintf(stderr, "\t*");
+	int curfont = 0;
+	for (const Setting *s = vc->setting.begin(); s != vc->setting.end(); s++)
+	    switch (s->op) {
+	      case Setting::FONT:
+		fprintf(stderr, " {F%d}", s->x);
+		curfont = s->x;
+		break;
+	      case Setting::SHOW:
+		fprintf(stderr, " %d", s->x);
+		if (curfont == 0)
+		    fprintf(stderr, "/%s", code_str(s->x));
+		break;
+	      case Setting::KERN:
+		fprintf(stderr, " <>");
+		break;
+	      case Setting::MOVE:
+		fprintf(stderr, " <%+d,%+d>", s->x, s->y);
+		break;
+	      case Setting::RULE:
+		fprintf(stderr, " [%d,%d]", s->x, s->y);
+		break;
+	      case Setting::PUSH:
+		fprintf(stderr, " (");
+		break;
+	      case Setting::POP:
+		fprintf(stderr, " )");
+		break;
+	      case Setting::SPECIAL:
+		fprintf(stderr, " S{%s}", s->s.c_str());
+		break;
+	    }
+	fprintf(stderr, "  ((%d/%s, %d/%s))\n", ch->built_in1, code_str(ch->built_in1), ch->built_in2, code_str(ch->built_in2));
+    }
+    for (const Ligature *l = ch->ligatures.begin(); l != ch->ligatures.end(); l++)
+	fprintf(stderr, "\t[%d/%s => %d/%s]%s\n", l->in2, code_str(l->in2), l->out, code_str(l->out), (_encoding[l->out].context_setting(c, l->in2) ? " [C]" : ""));
+#if 0
+    for (const Kern *k = ch->kerns.begin(); k != ch->kerns.end(); k++)
+	fprintf(stderr, "\t{%d/%s %+d}\n", k->in2, code_str(k->in2), k->kern);
+#endif
+}
+
 void
 Metrics::unparse() const
 {
-    for (Code c = 0; c < _encoding.size(); c++)
-	if (_encoding[c].glyph) {
-	    const Char &ch = _encoding[c];
-	    fprintf(stderr, "%4d/%s%s%s%s%s%s\n", c, code_str(c),
-		    (ch.flag(Char::LIVE) ? " [L]" : ""),
-		    (ch.flag(Char::BASE_LIVE) ? " [B]" : ""),
-		    (ch.flag(Char::CONTEXT_ONLY) ? " [C]" : ""),
-		    (ch.flag(Char::BUILT) ? " [!]" : ""),
-		    (ch.base_code >= 0 ? " <BC>" : ""));
-	    if (ch.base_code >= 0 && ch.base_code != c)
-		fprintf(stderr, "\tBASE %d/%s\n", ch.base_code, code_str(ch.base_code));
-	    if (const VirtualChar *vc = ch.virtual_char) {
-		fprintf(stderr, "\t*");
-		for (const Setting *s = vc->setting.begin(); s != vc->setting.end(); s++)
-		    switch (s->op) {
-		      case Setting::FONT:
-			fprintf(stderr, " {F%d}", s->x);
-			break;
-		      case Setting::SHOW:
-			fprintf(stderr, " %d/%s", s->x, code_str(s->x));
-			break;
-		      case Setting::KERN:
-			fprintf(stderr, " <>");
-			break;
-		      case Setting::MOVE:
-			fprintf(stderr, " <%+d,%+d>", s->x, s->y);
-			break;
-		      case Setting::RULE:
-			fprintf(stderr, " [%d,%d]", s->x, s->y);
-			break;
-		      case Setting::PUSH:
-			fprintf(stderr, " (");
-			break;
-		      case Setting::POP:
-			fprintf(stderr, " )");
-			break;
-		      case Setting::SPECIAL:
-			fprintf(stderr, " S{%s}", s->s.c_str());
-			break;
-		    }
-		fprintf(stderr, "  ((%d/%s, %d/%s))\n", ch.built_in1, code_str(ch.built_in1), ch.built_in2, code_str(ch.built_in2));
-	    }
-	    for (const Ligature *l = ch.ligatures.begin(); l != ch.ligatures.end(); l++)
-		fprintf(stderr, "\t[%d/%s => %d/%s]%s\n", l->in2, code_str(l->in2), l->out, code_str(l->out), (_encoding[l->out].context_setting(c, l->in2) ? " [C]" : ""));
-#if 0
-	    for (const Kern *k = ch.kerns.begin(); k != ch.kerns.end(); k++)
-		fprintf(stderr, "\t{%d/%s %+d}\n", k->in2, code_str(k->in2), k->kern);
-#endif
-	}
+    for (const Char *ch = _encoding.begin(); ch < _encoding.end(); ch++)
+	if (ch->glyph)
+	    unparse(ch);
 }
