@@ -1,6 +1,6 @@
 /* otftotfm.cc -- driver for translating OpenType fonts to TeX metrics
  *
- * Copyright (c) 2003-2006 Eddie Kohler
+ * Copyright (c) 2003-2007 Eddie Kohler
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -97,16 +97,17 @@ using namespace Efont;
 #define MATH_SPACING_OPT	338
 #define POSITION_OPT		339
 #define WARN_MISSING_OPT	340
+#define BASE_ENCODINGS_OPT	341
 
-#define AUTOMATIC_OPT		341
-#define FONT_NAME_OPT		342
-#define QUIET_OPT		343
-#define GLYPHLIST_OPT		344
-#define VENDOR_OPT		345
-#define TYPEFACE_OPT		346
-#define NOCREATE_OPT		347
-#define VERBOSE_OPT		348
-#define FORCE_OPT		349
+#define AUTOMATIC_OPT		342
+#define FONT_NAME_OPT		343
+#define QUIET_OPT		344
+#define GLYPHLIST_OPT		345
+#define VENDOR_OPT		346
+#define TYPEFACE_OPT		347
+#define NOCREATE_OPT		348
+#define VERBOSE_OPT		349
+#define FORCE_OPT		350
 
 #define VIRTUAL_OPT		351
 #define PL_OPT			352
@@ -145,6 +146,7 @@ static Clp_Option options[] = {
     { "subs-filter", 0, SUBS_FILTER_OPT, Clp_ArgString, 0 },
     { "encoding", 'e', ENCODING_OPT, Clp_ArgString, 0 },
     { "literal-encoding", 0, LITERAL_ENCODING_OPT, Clp_ArgString, 0 },
+    { "base-encodings", 0, BASE_ENCODINGS_OPT, Clp_ArgString, 0 },
     { "extend", 'E', EXTEND_OPT, Clp_ArgDouble, 0 },
     { "slant", 'S', SLANT_OPT, Clp_ArgDouble, 0 },
     { "letterspacing", 'L', LETTERSPACE_OPT, Clp_ArgInt, 0 },
@@ -220,6 +222,12 @@ hyphen hyphen =: endash ; endash hyphen =: emdash ; \
 quoteleft quoteleft =: quotedblleft ; \
 quoteright quoteright =: quotedblright ;";
 
+struct BaseEncoding {
+    String font_name;
+    String secondary;
+    DvipsEncoding encoding;
+};
+
 static const char *program_name;
 static String::Initializer initializer;
 static String current_time;
@@ -237,9 +245,10 @@ static HashMap<Efont::OpenType::Tag, GlyphFilter*> altselector_feature_filters(0
 
 static String font_name;
 static String encoding_file;
+static Vector<BaseEncoding *> base_encodings;
 static double extend;
 static double slant;
-static int letterspace;
+int letterspace;
 static double design_size;
 static double minimum_kern = 2.0;
 static double space_factor = 1.0;
@@ -321,6 +330,7 @@ Encoding options:\n\
       --coding-scheme=SCHEME   Set the output coding scheme to SCHEME.\n\
       --warn-missing           Warn about characters not supported by font.\n\
       --literal-encoding=FILE  Use DVIPS encoding FILE verbatim.\n\
+      --base-encodings=FILE    Output can refer to base fonts named in FILE.\n\
 \n");
     printf("\
 Automatic mode options:\n\
@@ -470,13 +480,18 @@ lig_context_str(int ctx)
     return (ctx == 0 ? "LIG" : (ctx < 0 ? "/LIG" : "LIG/"));
 }
 
+static double max_printed_real;
+
 static void
 fprint_real(FILE *f, const char *prefix, double value, double du, const char *suffix = ")\n")
 {
-    if (du == 1.)
+    if (du == 1.) {
 	fprintf(f, "%s R %g%s", prefix, value, suffix);
-    else
+	max_printed_real = std::max(max_printed_real, fabs(value));
+    } else {
         fprintf(f, "%s R %.4f%s", prefix, value * du, suffix);
+	max_printed_real = std::max(max_printed_real, fabs(value * du));
+    }
 }
 
 static String
@@ -499,9 +514,24 @@ font_slant(const FontInfo &finfo)
 }
 
 static void
-output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
-	  const FontInfo &finfo, bool vpl, FILE *f)
+output_pl(Metrics &metrics, const String &ps_name, int boundary_char,
+	  const FontInfo &finfo, bool vpl,
+	  const String &filename, ErrorHandler *errh)
 {
+    // create file
+    if (no_create) {
+	errh->message("would create %s", filename.c_str());
+	return;
+    }
+
+    if (verbose)
+	errh->message("creating %s", filename.c_str());
+    FILE *f = fopen(filename.c_str(), "w");
+    if (!f) {
+	errh->error("%s: %s", filename.c_str(), strerror(errno));
+	return;
+    }
+
     // XXX check DESIGNSIZE and DESIGNUNITS for correctness
 
     fprintf(f, "(COMMENT Created by '%s'%s)\n", invocation.c_str(), current_time.c_str());
@@ -518,6 +548,8 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 
     if (design_size <= 0)
 	design_size = get_design_size(finfo);
+    max_printed_real = 0;
+    
     fprintf(f, "(DESIGNSIZE R %.1f)\n"
 	    "(DESIGNUNITS R %d.0)\n"
 	    "(COMMENT DESIGNSIZE (1 em) IS IN POINTS)\n"
@@ -539,7 +571,7 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 
     if (char_bounds(bounds, width, finfo, font_xform, ' ')) {
 	// advance space width by letterspacing, scale by space_factor
-	double space_width = (width + letterspace) * space_factor;
+	double space_width = (width + (vpl ? letterspace : 0)) * space_factor;
 	fprint_real(f, "   (SPACE", space_width, du);
 	if (finfo.is_fixed_pitch()) {
 	    // fixed-pitch: no space stretch or shrink
@@ -563,14 +595,43 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
     if (boundary_char >= 0)
 	fprintf(f, "(BOUNDARYCHAR D %d)\n", boundary_char);
 
-    // write MAPFONT
-    if (vpl)
-	for (int i = 0; i < metrics.n_mapped_fonts(); i++) {
-	    String name = metrics.mapped_font_name(i);
+    // figure out font mapping
+    int mapped_font0 = 0;
+    Vector<int> font_mapping;
+    if (vpl) {
+	int vpl_first_font = (metrics.need_base() ? 0 : 1);
+	font_mapping.assign(metrics.n_mapped_fonts(), 0);
+	// how many times is each font used?
+	Vector<Setting> settings;
+	for (int i = 0; i < 256; i++)
+	    if (metrics.setting(i, settings)) {
+		int font_number = 0;
+		for (const Setting *s = settings.begin(); s < settings.end(); s++)
+		    if (s->op == Setting::SHOW)
+			font_mapping[font_number]++;
+		    else if (s->op == Setting::FONT)
+			font_number = (int) s->x;
+	    }
+	// make sure the most-used font is number 0
+	mapped_font0 = std::max_element(font_mapping.begin(), font_mapping.end()) - font_mapping.begin();
+	// prepare the mapping
+	for (int i = vpl_first_font; i < font_mapping.size(); i++)
+	    font_mapping[i] = i - vpl_first_font;
+	font_mapping[mapped_font0] = 0;
+	font_mapping[vpl_first_font] = mapped_font0 - vpl_first_font;
+	if (vpl_first_font != 0)
+	    font_mapping[0] = font_mapping.size() - 1;
+	// write MAPFONT
+	for (int i = 0; i < metrics.n_mapped_fonts() - vpl_first_font; i++) {
+	    int j = std::find(font_mapping.begin(), font_mapping.end(), i) - font_mapping.begin();
+	    String name = metrics.mapped_font_name(j);
 	    if (!name)
 		name = make_base_font_name(font_name);
 	    fprintf(f, "(MAPFONT D %d\n   (FONTNAME %s)\n   (FONTDSIZE R %.1f)\n   )\n", i, name.c_str(), design_size);
 	}
+    } else
+	for (int i = 0; i < metrics.n_mapped_fonts(); i++)
+	    font_mapping.push_back(i);
     
     // figure out the proper names and numbers for glyphs
     Vector<String> glyph_ids;
@@ -661,38 +722,50 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 	    sa.clear();
 	    push_stack.clear();
 	    CharstringBounds boundser(font_xform);
+	    int program_number = mapped_font0;
 	    const CharstringProgram *program = finfo.program();
 	    for (const Setting *s = settings.begin(); s < settings.end(); s++)
 		switch (s->op) {
 		    
 		  case Setting::SHOW:
-		    boundser.char_bounds(program->glyph_context(s->y));
+		    if (vpl || program == finfo.program())
+			boundser.char_bounds(program->glyph_context(s->y));
 		    // 3.Aug.2004 -- reported by Marco Kuhlmann: Don't use
 		    // glyph_ids[] array when looking at a different font.
-		    if (program == finfo.program())
+		    if (program_number == 0)
 			sa << "      (SETCHAR " << glyph_ids[s->x] << ')' << glyph_base_comments[s->x] << "\n";
 		    else
 			sa << "      (SETCHAR D " << s->x << ")\n";
 		    break;
 
-		  case Setting::MOVE:
-		    boundser.translate(s->x, s->y);
-		    if (s->x)
-			sa << "      (MOVERIGHT R " << real_string(s->x, du) << ")\n";
-		    if (s->y)
-			sa << "      (MOVEUP R " << real_string(s->y, du) << ")\n";
-		    break;
+		  case Setting::MOVE: {
+		      int x = 0, y = 0;
+		      while (s+1 < settings.end() && s[1].op == Setting::MOVE)
+			  x += s->x, y += s->y, s++;
+		      if (vpl)
+			  boundser.translate(s->x + x, s->y + y);
+		      if (s->x + x)
+			  sa << "      (MOVERIGHT R " << real_string(s->x + x, du) << ")\n";
+		      if (s->y + y)
+			  sa << "      (MOVEUP R " << real_string(s->y + y, du) << ")\n";
+		      break;
+		  }
 
 		  case Setting::RULE:
-		    boundser.mark(Point(0, 0));
-		    boundser.mark(Point(s->x, s->y));
-		    boundser.translate(s->x, 0);
+		    if (vpl) {
+			boundser.mark(Point(0, 0));
+			boundser.mark(Point(s->x, s->y));
+			boundser.translate(s->x, 0);
+		    }
 		    sa << "      (SETRULE R " << real_string(s->y, du) << " R " << real_string(s->x, du) << ")\n";
 		    break;
 
 		  case Setting::FONT:
-		    program = metrics.mapped_font((int) s->x);
-		    sa << "      (SELECTFONT D " << (int) s->x << ")\n";
+		    if ((int) s->x != program_number) {
+			program = metrics.mapped_font((int) s->x);
+			program_number = (int) s->x;
+			sa << "      (SELECTFONT D " << font_mapping[program_number] << ")\n";
+		    }
 		    break;
 
 		  case Setting::PUSH:
@@ -703,7 +776,8 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 		  case Setting::POP: {
 		      assert(push_stack.size());
 		      Point p = push_stack.back() - boundser.transform(Point(0, 0));
-		      boundser.translate(p.x, p.y);
+		      if (vpl)
+			  boundser.translate(p.x, p.y);
 		      push_stack.pop_back();
 		      sa << "      (POP)\n";
 		      break;
@@ -744,22 +818,18 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 		fprintf(f, "   (MAP\n%s      )\n", sa.c_str());
 	    fprintf(f, "   )\n");
 	}
-}
 
-static void
-output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
-	  const FontInfo &finfo, bool vpl, String filename, ErrorHandler *errh)
-{
-    if (no_create)
-	errh->message("would create %s", filename.c_str());
-    else {
+    // at last, close the file
+    fclose(f);
+    
+    // Did we print a number too big for TeX to handle?  If so, try again.
+    if (max_printed_real >= 2047) {
+	if (metrics.design_units() <= 1)
+	    errh->fatal("This font appears to be broken.  It has characters so big that the PL format\ncannot represent them.");
+	metrics.set_design_units(metrics.design_units() > 200 ? metrics.design_units() - 250 : 1);
 	if (verbose)
-	    errh->message("creating %s", filename.c_str());
-	if (FILE *f = fopen(filename.c_str(), "w")) {
-	    output_pl(metrics, ps_name, boundary_char, finfo, vpl, f);
-	    fclose(f);
-	} else
-	    errh->error("%s: %s", filename.c_str(), strerror(errno));
+	    errh->message("the font's metrics overflow the limits of PL files\n(reducing DESIGNUNITS to %d and trying again)", metrics.design_units());
+	output_pl(metrics, ps_name, boundary_char, finfo, vpl, filename, errh);
     }
 }
 
@@ -906,7 +976,7 @@ write_encoding_file(String &filename, const String &encoding_name,
     return 0;
 }
 
-static void
+static bool
 output_encoding(const Metrics &metrics,
 		const Vector<PermString> &glyph_names,
 		ErrorHandler *errh)
@@ -915,7 +985,9 @@ output_encoding(const Metrics &metrics,
 
     // collect encoding data
     Vector<Metrics::Glyph> glyphs;
-    metrics.base_glyphs(glyphs);
+    if (!metrics.base_glyphs(glyphs, 256))
+	return false;
+    
     StringAccum sa;
     for (int i = 0; i < 256; i++) {
 	if ((i & 0xF) == 0)
@@ -947,7 +1019,7 @@ output_encoding(const Metrics &metrics,
 
     // exit if we're not responsible for generating an encoding
     if (!(output_flags & G_ENCODING))
-	return;
+	return true;
 
     // put encoding block in a StringAccum
     // 3.Jun.2003: stick command line definition at the end of the encoding,
@@ -992,29 +1064,26 @@ output_encoding(const Metrics &metrics,
 	fwrite(contents.data(), 1, contents.length(), stdout);
     else if (write_encoding_file(out_encoding_file, out_encoding_name, contents, errh) == 1)
 	update_odir(O_ENCODING, out_encoding_file, errh);
+    return true;
 }
 
 static void
-output_tfm(const Metrics &metrics, const String &ps_name, int boundary_char,
+output_tfm(Metrics &metrics, const String &ps_name, int boundary_char,
 	   const FontInfo &finfo, String tfm_filename, String vf_filename,
 	   ErrorHandler *errh)
 {
     String pl_filename;
-    int pl_fd = temporary_file(pl_filename, errh);
-    if (pl_fd < 0)
-	return;
-
     bool vpl = vf_filename;
 
     if (no_create) {
 	errh->message("would write %s to temporary file", (vpl ? "VPL" : "PL"));
 	pl_filename = "<temporary>";
     } else {
-	if (verbose)
-	    errh->message("writing %s to temporary file", (vpl ? "VPL" : "PL"));
-	FILE *f = fdopen(pl_fd, "w");
-	output_pl(metrics, ps_name, boundary_char, finfo, vpl, f);
-	fclose(f);
+	int pl_fd = temporary_file(pl_filename, errh);
+	if (pl_fd < 0)
+	    return;
+	output_pl(metrics, ps_name, boundary_char, finfo, vpl, pl_filename, errh);
+	close(pl_fd);
     }
 
     StringAccum command;
@@ -1080,9 +1149,13 @@ output_metrics(Metrics &metrics, const String &ps_name, int boundary_char,
 	    output_pl(metrics, ps_name, boundary_char, finfo, true, outfile, errh);
 	    update_odir(O_VPL, outfile, errh);
 	}
-	metrics.make_base(257);
     }
 
+    // quit if no base needed
+    metrics.make_base(257);
+    if (!metrics.need_base())
+	return;
+    
     // output metrics
     double save_minimum_kern = minimum_kern;
     if (need_virtual)
@@ -1296,22 +1369,19 @@ do_math_spacing(Metrics &metrics, const FontInfo &finfo,
     
     int bounds[4], width;
     
-    for (int code = 0; code < 256; code++) {
-	int g = metrics.glyph(code);
-	if (g != 0 && g != Metrics::VIRTUAL_GLYPH && code != boundary_char) {
-	    if (char_bounds(bounds, width, finfo, font_xform, code)) {
-		int left_sb = (bounds[0] < 0 ? -bounds[0] : 0);
-		metrics.add_single_positioning(code, left_sb, 0, left_sb);
-		
-		if (skew_char >= 0) {
-		    double virtual_height = (bounds[3] > x_height ? bounds[3] :x_height) - 0.5 * x_height;
-		    int right_sb = (bounds[2] > width ? bounds[2] - width : 0);
-		    int skew = (int)(slant * virtual_height + left_sb - right_sb);
-		    metrics.add_kern(code, skew_char, skew);
-		}
+    for (int code = 0; code < metrics.encoding_size(); code++)
+	if (metrics.was_base_glyph(code) && code != boundary_char
+	    && char_bounds(bounds, width, finfo, font_xform, code)) {
+	    int left_sb = (bounds[0] < 0 ? -bounds[0] : 0);
+	    metrics.add_single_positioning(code, left_sb, 0, left_sb);
+	    
+	    if (skew_char >= 0 && code < 256) {
+		double virtual_height = (bounds[3] > x_height ? bounds[3] :x_height) - 0.5 * x_height;
+		int right_sb = (bounds[2] > width ? bounds[2] - width : 0);
+		int skew = (int)(slant * virtual_height + left_sb - right_sb);
+		metrics.add_kern(code, skew_char, skew);
 	    }
 	}
-    }
 }
 
 static void
@@ -1390,22 +1460,29 @@ do_file(const String &otf_filename, const OpenType::Font &otf,
     dvipsenc.apply_ligkern_kern(metrics, errh);
     dvipsenc.apply_position(metrics, errh);
 
+    // use prespecified raw fonts
+    for (BaseEncoding **be = base_encodings.begin(); be != base_encodings.end(); be++)
+	if (!(*be)->secondary) {
+	    Vector<int> mapp;
+	    (*be)->encoding.make_base_mappings(mapp, finfo);
+	    metrics.apply_base_encoding((*be)->font_name, (*be)->encoding, mapp);
+	}
+
     // remove extra characters
     metrics.cut_encoding(257);
     //metrics.unparse();
     int boundary_char = dvipsenc.boundary_char();
 
     // apply letterspacing, if any
-    if (letterspace) {
-	for (int code = 0; code < 256; code++) {
-	    int g = metrics.glyph(code);
-	    if (g != 0 && g != Metrics::VIRTUAL_GLYPH && code != boundary_char) {
+    if (letterspace)
+	for (int code = 0; code < metrics.encoding_size(); code++)
+	    if (metrics.was_base_glyph(code) && code != boundary_char) {
 		metrics.add_single_positioning(code, letterspace / 2, 0, letterspace);
-		metrics.add_kern(code, 256, -letterspace / 2);
-		metrics.add_kern(256, code, -letterspace / 2);
+		if (code < 256) {
+		    metrics.add_kern(code, 256, -letterspace / 2);
+		    metrics.add_kern(256, code, -letterspace / 2);
+		}
 	    }
-	}
-    }
 
     // apply math letterspacing, if any
     if (math_spacing)
@@ -1471,6 +1548,16 @@ do_file(const String &otf_filename, const OpenType::Font &otf,
 }
 
 
+String
+installed_metrics_font_name(const String &base_font_name, const String &secondary)
+{
+    for (BaseEncoding **be = base_encodings.begin(); be != base_encodings.end(); be++)
+	if ((*be)->secondary == secondary && ::font_name == base_font_name)
+	    return (*be)->font_name;
+    return String();
+}
+
+
 extern "C" {
 static int
 clp_parse_char(Clp_Parser *clp, const char *arg, int complain, void *)
@@ -1490,6 +1577,63 @@ clp_parse_char(Clp_Parser *clp, const char *arg, int complain, void *)
 }
 }
 
+static void
+parse_base_encodings(const String &filename, ErrorHandler *errh)
+{
+    String str = read_file(filename, errh, true);
+    String print_filename = (filename == "-" ? "<stdin>" : filename) + ":";
+    int lineno = 1;
+    str.c_str();
+    const char *s_end = str.end();
+    for (const char *s = str.begin(); s != s_end; ) {
+	while (s != s_end && isspace(*s) && *s != '\n' && *s != '\r')
+	    s++;
+	// skip comments and blank lines
+	if (s != s_end && *s != '%' && *s != '#' && *s != '\n' && *s != '\r') {
+	    const char *w1 = s;
+	    while (s != s_end && !isspace(*s))
+		s++;
+	    BaseEncoding *be = new BaseEncoding;
+	    be->font_name = str.substring(w1, s);
+	    while (s != s_end && isspace(*s) && *s != '\n' && *s != '\r')
+		s++;
+	    const char *w2 = s;
+	    while (s != s_end && !isspace(*s))
+		s++;
+	    String efile = str.substring(w2, s);
+	    while (s != s_end && isspace(*s) && *s != '\n' && *s != '\r')
+		s++;
+	    if (s != s_end && !isspace(*s) && *s != '%' && *s != '#') {
+		const char *w3 = s;
+		while (s != s_end && !isspace(*s))
+		    s++;
+		be->secondary = str.substring(w3, s);
+	    }
+	    LandmarkErrorHandler lerrh(errh, print_filename + String(lineno));
+	    int before = lerrh.nerrors();
+	    if (be->secondary)
+		/* encoding ignored */;
+	    else if (!efile)
+		lerrh.error("missing encoding name");
+	    else if (String path = locate_encoding(efile, errh))
+		be->encoding.parse(path, true, true, &lerrh);
+	    else
+		lerrh.error("encoding '%s' not found", efile.c_str());
+	    if (lerrh.nerrors() == before)
+		base_encodings.push_back(be);
+	    else
+		delete be;
+	}
+	while (s != s_end && *s != '\n' && *s != '\r')
+	    s++;
+	if (s != s_end && *s == '\r')
+	    s++;
+	if (s != s_end && *s == '\n')
+	    s++;
+	lineno++;
+    }
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -1521,6 +1665,7 @@ main(int argc, char *argv[])
     Vector<String> ligkern;
     Vector<String> pos;
     Vector<String> unicoding;
+    Vector<String> base_encoding_files;
     bool no_ecommand = false, default_ligkern = true;
     int warn_missing = -1;
     String codingscheme;
@@ -1610,6 +1755,10 @@ main(int argc, char *argv[])
 	    literal_encoding = true;
 	    break;
 
+	  case BASE_ENCODINGS_OPT:
+	    base_encoding_files.push_back(clp->arg);
+	    break;
+	    
 	  case EXTEND_OPT:
 	    if (extend)
 		usage_error(errh, "extend value specified twice");
@@ -1841,7 +1990,7 @@ main(int argc, char *argv[])
 
 	  case VERSION_OPT:
 	    printf("otftotfm (LCDF typetools) %s\n", VERSION);
-	    printf("Copyright (C) 2002-2006 Eddie Kohler\n\
+	    printf("Copyright (C) 2002-2007 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
@@ -1943,6 +2092,10 @@ particular purpose.\n");
 	    if (String s = read_file(*g, errh, true))
 		DvipsEncoding::add_glyphlist(s);
 
+	// read base encodings
+	for (String *s = base_encoding_files.begin(); s < base_encoding_files.end(); s++)
+	    parse_base_encodings(*s, errh);
+
 	// read encoding
 	DvipsEncoding dvipsenc;
 	if (encoding_file) {
